/*
**　高精度ＡＤコンバータ　ＡＤ７７０５
**　１６ビットΣΔ型、リファレンス電圧２．５Ｖ　差動入力
**  発電鍋で炊飯、熱電対・サーミスタ温度計測、Ｎｉ－ＭＨ充電
*/

//　液晶モジュール用ヘッダファイル
#include <LiquidCrystal.h>
//　EEPROM用ヘッダファイル
#include <EEPROM.h>

// １０３ＪＴサーミスタ温度テーブル
int jt_tablelen = 15;
unsigned int jt_adv[] = { 0, 549, 1860, 3172, 4481, 5793, 7102, 8420, 9727, 11038, 12352, 13661, 14975, 16286, 17599 };
float jt_tmp[] = { 0.0, 1.68, 4.21, 6.61, 8.87, 11.64, 14.93, 18.05, 21.43, 25.51, 30.30, 36.44, 43.35, 52.69, 66.42 };

// 移動平均バッファと炊飯フラグ
float mvavg[10] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
int mvptr = 0;
int suihan = 0;

//　液晶モジュールのピン設定
LiquidCrystal lcd(8, 9, 4, 5, 6, 7);

// ＡＤ７７０５とのインターフェイス（入出力はArdionoのポートの方向）
int DRDY = 2;  // データレディ（入力）
int CS = 3;    //　チップセレクト（出力）
int DIN = 11;  //　データ（出力）
int DOUT = 12;  //　データ（入力）
int SCLK = 13;   //　シリアルクロック（出力）

// ＡＤ７７０５チャンネル番号
int AD7705_CH = 0;

// 圧電ブザー
int PBUZZER = 10;

//　ａｄ７７０５への１バイト送信（ＳＣＬＫとＤＩＮのみ駆動）
void send_AD7705( unsigned char d ) {
  unsigned char i;
  for( i = 0; i < 8; i ++ ) {  //　８ビット分繰り返し
    digitalWrite( SCLK, LOW ); //　クロック線をＬＯＷレベルにする
    if( d & 0x80 ) digitalWrite( DIN, HIGH ); //　データビットを準備
    else           digitalWrite( DIN, LOW );
    digitalWrite( SCLK, HIGH );  //　クロック線をＨＩＧＨレベルに戻す
    d *= 2;  //　左シフト
  }
}

//　ＡＤ７７０５から１バイト受信（ＳＣＬＫのみ駆動）
unsigned char recv_AD7705() {
  unsigned char i, d;
  d = 0;
  for( i = 0; i < 8; i ++ ) {
    digitalWrite( SCLK, LOW ); //　クロック線をＬＯＷレベルにする
    d *= 2; //左シフト  
    if( HIGH == digitalRead( DOUT ) ) d += 1;  // HIGHならビットを立てる
    digitalWrite( SCLK, HIGH );  //　クロック線をＨＩＧＨレベルに戻す
  }
  return d;
}

// ＡＤ７７０５のシリアルポートを初期化
void reset_AD7705() {
  digitalWrite( CS, LOW );  //　アクセス開始
  send_AD7705( 0xff );    //　最低３２ビットのＨレベルを送信することで
  send_AD7705( 0xff );    //　ＡＤ７７０５のシリアルポートをリセットできる
  send_AD7705( 0xff );
  send_AD7705( 0xff );
  send_AD7705( 0xff );
  digitalWrite( CS, HIGH );  //　リセット終了
  delay( 1 );
}

// ＡＤ７７０５の８ビットレジスタへの書込み
void write8_AD7705( unsigned char com, unsigned char dat ) {
  digitalWrite( CS, LOW );  //　アクセス開始
  send_AD7705( com );    //　コミュニケーションレジスタ
  digitalWrite( CS, HIGH ); 
  delayMicroseconds( 10 );
  digitalWrite( CS, LOW );  //　アクセス開始
  send_AD7705( dat );    // レジスタへのデータ
  digitalWrite( CS, HIGH ); 
  delay( 1 );
}
 
//　ＡＤ７７０５の初期化
//  ch = channel 0 or 1 )
//  gain = gain code( 0..7 )
void setup_AD7705( int ch, int gain ) {
//　１．ポートの向きを設定
  pinMode( DRDY, INPUT );
  pinMode( CS, OUTPUT );
  pinMode( DIN, OUTPUT );
  pinMode( DOUT, INPUT );
  pinMode( SCLK, OUTPUT );
//　２．出力ポートはＨレベルにしておく
  digitalWrite( CS, HIGH );
  digitalWrite( DIN, HIGH );
  digitalWrite( SCLK, HIGH );
// チャンネル番号初期化
  AD7705_CH = 0;
// ＡＤ７７０５シリアルポート初期化
  delay( 10 );    // １０ｍＳ待ち
  reset_AD7705();
// 初期化:クロック設定
  write8_AD7705( 0x20, 0x08 );
  // Comunication reg.
  //   .010.... select clock reg.
  //   ....0... write
  //   ,,,,,0.. normal operation
  //   ......00 channel 0
  // Clock reg.
  //   000..... always zerp
  //   ...0.... clock enable
  //   ....10.. crystal 4.9152MHz
  //   ......00 data update rate 20Hz
  // 電源周波数が６０Ｈｚの地域では０ｘ０８、５０Ｈｚの地域では０ｘ０９
// 初期化:セットアップレジスタ設定による自動校正
  write8_AD7705( 0x10 + ch, 0x40 + 0x08 * gain );
  // Comunication reg.
  //   .001.... setup reg.
  //   ....0... write
  //   .....0.. normal operation
  //   ......00 channel
  // Setup reg.
  //   01...... self calibration
  //   ..***... PGA
  //   .....0.. bipoler
  //   ......0. onchip buffer off
  //   .......0 normal operation
  while( LOW == digitalRead( DRDY ) ) {  // 校正終了待ち
    delay( 1 );
  }
}

// ＡＤ７７０５からのデータ読み出し（変換が終わるまで待ってデータを読み出す）
unsigned int conv_AD7705( int ch ) {
  unsigned int d;
  unsigned char h, l;

  while( LOW == digitalRead( DRDY ) ) {  // 変換終了待ち
    delay( 1 );
  }
  digitalWrite( CS, LOW );  //　アクセス開始
  if( 0 != ch ) send_AD7705( 0x39 );    //　データレジスタ読み出しを選択
  else          send_AD7705( 0x38 );
  digitalWrite( CS, HIGH );
  delayMicroseconds( 10 );
  digitalWrite( CS, LOW );  //　アクセス開始
  h = recv_AD7705();        //　上位バイト
  l = recv_AD7705();        //　下位バイト
  digitalWrite( CS, HIGH );
  d = ( h << 8 ) + l;
  return d;
}

//　ＡＤコンバータの値から温度へ変換（１０３ＪＴサーミスタ）
float adv2tmp_JT( unsigned int v ) {
  unsigned int i, dc;
  float dt;

  if( 53000 < v ) return 0.0;  // ０℃以下なら計算せず０を返す
  v = 53000 - v;  // サーミスタは温度係数が負
  for( i = 0; i < jt_tablelen; i ++ ) {
    if( jt_adv[ i ] >= v ) break;
  }
  if( jt_tablelen == i ) return 99.9; // テーブルからはみ出したらエラー
  --i;
  // 折れ線近似
  dc = jt_adv[ i + 1 ] - jt_adv[ i ];
  dt = jt_tmp[ i + 1 ] - jt_tmp[ i ];
  v -= jt_adv[ i ];
  return jt_tmp[ i ] + ( float )v / ( float ) dc * dt; 
}

// ＡＤコンバータの値から温度差への変換（Ｋ熱電対）
float adv2tmp_K( unsigned int v ) {
  signed int a;
  a = ( int )( v - 32768);
  return ( float )a / 68.27;
}

// 圧電ブザーを鳴らす
void beep() {
  int i;

  for( i = 0; i < 2000; i ++ ) {
    digitalWrite( PBUZZER, HIGH );
    delayMicroseconds( 100 );
    digitalWrite( PBUZZER, LOW );
    delayMicroseconds( 100 );
  }
}

// 移動平均と炊飯フラグ処理
int mv_avarage( float t ) {
  int i, a;
  // 移動平均バッファへ温度を格納
  mvavg[ mvptr ] = t;
  mvptr ++;
  if( 10 <= mvptr ) mvptr = 0; 
  // バッファ内の平均値を求める
  a = 0.0;
  for( i = 0; i < 10; i ++ ) a += mvavg[ i ];
  a /= 10.0;
  // 炊飯フラグが０で温度が１０４度を超えたら加熱停止
  if( 0 == suihan && 104.0 < a ) {
    suihan = 1;
    beep();
    return 1;
  }
  // 炊飯フラグが１で温度が１００度未満になったらむらし終了
  if( 1 == suihan && 100.0 > a ) {
    suihan = 2;
    beep();
    return 2;
  }
  return 0;
}



//　初期化のために最初に１回だけ呼ばれる関数
void setup() {
  int i;

  lcd.begin(16, 2);        //　液晶モジュールの初期化
  Serial.begin( 9600 );   //　シリアルポートの初期化
  setup_AD7705( 0, 0 );    //　ＡＤ７７０５の初期化
  delay( 100 );            //　安定待ち

// 圧電ブザー初期化
  pinMode( PBUZZER, OUTPUT );
  digitalWrite( PBUZZER, LOW );
}

//　動作時に繰り返し呼ばれる関数
void loop() {
  long int d0, d1;
  int i;
  float t;

  reset_AD7705();
  setup_AD7705( 0, 7 );
  for( i = 0; i < 5; i ++ ) d0 = conv_AD7705( 0 );  
  delay( 500 );

  reset_AD7705();
  setup_AD7705( 1, 0 ); 
  for( i = 0; i < 5; i ++ ) d1 = conv_AD7705( 1 );

  t = adv2tmp_JT( d0 );
  lcd.setCursor( 0, 0 );
  lcd.print( "TH=" );
  lcd.print( t );
  lcd.print( " " );
  
  t += adv2tmp_K( d1 );  
  lcd.setCursor( 0, 1 );
  lcd.print( "TC=" );
  lcd.print( t );
  lcd.print( " " );

  i = mv_avarage( t );
  lcd.setCursor( 10, 0 );
  if( 1 == i ) lcd.print( "STOP!" );
  if( 2 == i ) lcd.print( "READY!" );

  Serial.print( d0 );
  Serial.print( "," );
  Serial.print( d1 );
  Serial.print( "," );
  Serial.println( t );
    
  delay( 500 );
}

